/*
** Module   : PersistenceFactory.java
** Abstract : Factory for Persistence objects
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------------Description-------------------------
** 001 ECF 20071018   @35570 Created initial version. Factory for
**                           Persistence objects. Can create local and
**                           remote versions.
** 002 ECF 20080611   @38698 Disallow creation of a Persistence instance
**                           for a dirty database. In order to prevent
**                           programming errors, IllegalArgumentException
**                           is thrown if this is attempted.
** 003 ECF 20131010          Minor documentation update.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import org.hibernate.dialect.*;
import com.goldencode.p2j.persist.remote.*;

/**
 * Factory for {@link Persistence} objects.  Manages a cache of
 * <code>Persistence</code> objects, keyed by {@link Database}.  When an
 * instance is requested via one of the <code>getInstance()</code> method
 * variants, the cache is checked for an existing instance.  If none exists,
 * the factory determines what type of <code>Persistence</code> implementation
 * is necessary for a particular database, instantiates it, initializes it,
 * caches it, and returns it.
 * <p>
 * Currently, this factory knows how to generate instances of:
 * <ul>
 *   <li>{@link Persistence} - used for locally managed databases.
 *   <li>{@link com.goldencode.p2j.persist.remote.RemotePersistence
 *       RemotePersistence} - used for remotely managed databases.
 * </ul>
 * <p>
 * Persistence objects can be removed from the cache by other persistence
 * framework classes via the {@link #remove(Database)} method.
 * 
 * @author ECF
 */
public final class PersistenceFactory
{
   /** Cache of <code>Persistence</code> instances by database */
   private static final Map<Database, Persistence> cache =
      new HashMap<Database, Persistence>();
   
   /**
    * Get the {@link Persistence} instance associated with the given, physical database name,
    * creating it first if necessary. Assumes the name specifies a primary database. Newly
    * created instances are stored in a static cache.
    *
    * @param   name
    *          Name of physical database with which the instance is
    *          permanently associated.
    *          
    * @return  Persistence object associated with the given database.
    */
   public static Persistence getInstance(String name)
   {
      return getInstance(new Database(name));
   }
   
   /**
    * Get the instance of this class associated with the given, physical database, creating it
    * first if necessary. Newly created instances are stored in a static cache.
    *
    * @param   database
    *          Database with which the instance is permanently associated.
    *          
    * @return  Persistence object associated with the given database.
    * 
    * @throws  IllegalArgumentException
    *          if <code>database</code> represents a dirty database.
    */
   public static Persistence getInstance(Database database)
   {
      if (database.isDirty())
      {
         throw new IllegalArgumentException(
            "Cannot create a Persistence instance for dirty database '" +
            database +
            "'");
      }
      
      synchronized (cache)
      {
         Persistence instance = cache.get(database);
         if (instance == null)
         {
            Dialect dialect = DatabaseManager.getDialect(database);
            if (dialect == null)
            {
               throw new IllegalArgumentException(
                  "Unable to determine database dialect for database " +
                  database);
            }
            
            instance = (database.isLocal()
                        ? new Persistence(database, dialect)
                        : new RemotePersistence(database, dialect));
            instance.initializeInstance();
            
            cache.put(database, instance);
         }
         
         return instance;
      }
   }
   
   /**
    * Remove the <code>Persistence</code> instance associated with the given
    * database from the global cache.
    * <p>
    * This method should be called with caution.  It is provided primarily to
    * support transient databases, which are connected explicitly, used for a
    * time by one or more users, then disconnected.
    * 
    * @param   database
    *          Database with which the instance is associated.
    * 
    * @return  Instance of <code>Persistence</code> which was removed
    */
   static Persistence remove(Database database)
   {
      synchronized (cache)
      {
         return cache.remove(database);
      }
   }
}
