/*
** Module   : TableRelation.java
** Abstract : Representation of a database relation between two tables
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050325   @20982 Created initial version. Represents a natural
**                           relation between two database tables in a
**                           Progress schema.
** 002 ECF 20050616   @21535 Reworked to support notion of the inverse of
**                           a relation. Added methods to report databases
**                           affected by the join and to report inverse
**                           join type.
** 003 ECF 20070628   @35343 Minor optimization. Replaced StringBuffer
**                           with StringBuilder.
** 004 ECF 20081216   @40939 Fixed normalizedTableName(). Temp table
**                           database names were the internal versions
**                           ("@temp_db" and "@work_db") used by the
**                           SchemaDictionary, but they need to be the
**                           standard "_temp" used everywhere else.
** 005 GES 20090518   @42403 Import change.
** 006 ECF 20131013          Implemented generics.
*/

package com.goldencode.p2j.schema;

import java.util.*;
import com.goldencode.ast.*;

/**
 * A representation of a <em>natural relation</em> between two Progress
 * tables. A natural relation is a one defined by the following rules:
 * <ul>
 * <li>the tables must have between them one or more field with a common
 *     name and data type;
 * <li>the common field(s) must participate in a unique index in at least
 *     one of the tables;
 * <li>only one such relationship between the two tables is allowed.
 * </ul>
 * The nature of this relation ultimately is used to define an analogous
 * relation between a set of tables in a traditional, relational database.
 * In such a database, the relation is defined by a unique, primary key in
 * one table, referenced by a (not necessarily unique) foreign key in the
 * other.
 */
class TableRelation
implements Comparable<TableRelation>,
           SchemaParserTokenTypes
{
   /** AST of left-hand table in a join */
   private Aast left = null;
   
   /** AST of right-hand table in a join */
   private Aast right = null;
   
   /** AST of table which represents the primary key end of the relation */
   private Aast primary = null;
   
   /** Normalized left table name */
   private String leftName = null;
   
   /** Normalized right table name */
   private String rightName = null;
   
   /** Names of common fields which define the relation */
   private Set<String> key = null;
   
   /** Set of (lowercase) names of databases affected by this join */
   private Set<String> databases = new HashSet<String>();
   
   /** Type of relation:  many-to-one or one-to-one */
   private int type = -1;
   
   /**
    * Construct a <code>Join</code>, given a left table, right table,
    * optionally the table which represents the primary key end of the
    * relation, and the set of field names which define the join.
    *
    * @param   left
    *          Left table. May not be <code>null</code>.
    * @param   right
    *          Right table. May not be <code>null</code>.
    * @param   primary
    *          Table which represents the primary key end of the join.
    *          May be <code>null</code>.
    * @param   key
    *          Set of names of the fields which comprise the join.
    * @param   type
    *          Type of relation. If this object represents a primary/foreign
    *          key relation, type is always described from the foreign point
    *          of view (i.e., <code>MANY_TO_ONE</code>). Otherwise, type must
    *          be <code>ONE_TO_ONE</code>.
    *
    * @throws  NullPointerException
    *          if either <code>left</code> or <code>right</code> is
    *          <code>null</code>.
    * @throws  IllegalArgumentException
    *          if <code>type</code> is not a valid relation type.
    */
   TableRelation(Aast left, Aast right, Aast primary, Set<String> key, int type)
   {
      if (left == null || right == null)
      {
         throw new NullPointerException();
      }
      
      switch (type)
      {
         case MANY_TO_ONE:
         case ONE_TO_ONE:
            break;
         default:
            throw new IllegalArgumentException("Invalid relation type: " + type);
      }
      
      this.primary = primary;
      this.key = key;
      
      if (primary != right)
      {
         this.left = left;
         this.right = right;
      }
      else
      {
         this.left = right;
         this.right = left;
      }
      
      leftName = normalizedTableName(this.left);
      rightName = normalizedTableName(this.right);
      
      databases.add(this.left.getParent().getText().toLowerCase());
      databases.add(this.right.getParent().getText().toLowerCase());
      
      this.type = type;
   }
   
   /**
    * Composes a normalized table name, which is the fully qualified table
    * name in all lowercase, in the form:
    * <code>&lt;database_name&gt;&lt;table_name&gt;</code>.
    *
    * @param   table
    *          AST which defines table.
    *
    * @return  Normalized table name as described above.
    */
   static String normalizedTableName(Aast table)
   {
      String dbName = table.getParent().getText();
      if (SchemaDictionary.TEMP_DB.equals(dbName) || SchemaDictionary.WORK_DB.equals(dbName))
      {
         dbName = SchemaDictionary.TEMP_TABLE_DB;
      }
      
      StringBuilder buf = new StringBuilder();
      buf.append(dbName).append('.').append(table.getText());
         
      return (buf.toString().toLowerCase());
   }
   
   /**
    * A simplified implementation which considers two joins equal if they
    * both contain the same pair of tables, regardless of the left-ness or
    * right-ness of either table in each pair. The premise is that any
    * given pair of tables can only have one joinable relation between
    * them, regardless of the side of the join they are on. Note that this
    * premise does not consider the result set generated by a join (which
    * could vary for an outer join depending on the left-ness or right-ness
    * of the tables), only the nature of the relation between the tables.
    * <p>
    * Two tables are considered equal if their full names match.
    *
    * @param   o
    *          Object with which to compare this <code>Join</code>
    *          instance for equality.
    *
    * @return  <code>true</code> if the joins are equal according to the
    *          above rules, else <code>false</code>.
    */
   public boolean equals(Object o)
   {
      if (!(o instanceof TableRelation))
      {
         return false;
      }
      
      TableRelation rel = (TableRelation) o;
      
      String l1 = leftName;
      String l2 = rel.leftName;
      String r1 = rightName;
      String r2 = rel.rightName;
      
      return ((l1.equals(l2) && r1.equals(r2)) || (l1.equals(r2) && r1.equals(l2)));
   }
   
   /**
    * Override the default implementation of this method to be consistent
    * with the simplified implementation of {@link #equals}. It is
    * intentional that we get the same hash code regardless of which table
    * is "left" and which is "right".
    *
    * @return  Hash code for this object.
    */
   public int hashCode()
   {
      return leftName.hashCode() + rightName.hashCode();
   }
   
   /**
    * Compares this instance to another <code>TableRelation</code> instance
    * to determine how they are logically sorted:
    * <ol>
    * <li>relations which compare as equivalent by {@link #equals} always
    *     compare as equivalent by this method;
    * <li>relations which define a primary key are considered "greater than"
    *     those which do not;  if this test is inconclusive...
    * <li>the names of the left-hand tables of each relation are compared
    *     lexicographically;  if this test is inconclusive...
    * <li>the names of the right-hand tables of each relation are compared
    *     lexicographically.
    * </ol>
    *
    * @param   other
    *          Another instance of this class to compare against this one.
    *
    * @return  &lt; 0 if <code>this</code> is "less than" <code>o</code>;
    *          &gt; 0 if <code>this</code> is "greater than" <code>o</code>;
    *          0 if both instances compare as equivalent by the above rules.
    */
   public int compareTo(TableRelation other)
   {
      if (this.equals(other))
      {
         return 0;
      }
      
      int res = 0;
      if (primary == null && other.primary != null)
      {
         res = -1;
      }
      else if (primary != null && other.primary == null)
      {
         res = 1;
      }
      
      if (res == 0)
      {
         res = left.getText().compareTo(other.left.getText());
      }
      
      if (res == 0)
      {
         res = right.getText().compareTo(other.right.getText());
      }
      
      return res;
   }
   
   /**
    * Compose and return a string representation of this object, suitable
    * for simple reporting and debug purposes.
    *
    * @return  String representation of this object's state.
    */
   public String toString()
   {
      StringBuilder buf = new StringBuilder();
      
      buf.append(left.getParent().getText()).append('.').append(left.getText());
      
      if (primary == left)
      {
         buf.append(" <---- ");
      }
      else if (primary == right)
      {
         buf.append(" ----> ");
      }
      else
      {
         buf.append(" <---> ");
      }
      
      buf.append(right.getParent().getText()).append('.').append(right.getText());
      
      switch (type)
      {
         case MANY_TO_ONE:
            buf.append(" (many-to-one on ");
            break;
         case ONE_TO_ONE:
            buf.append(" (one-to-one on ");
            break;
      }
      
      buf.append(key).append(')');
      
      return buf.toString();
   }
   
   /**
    * Get the left-hand table of the join defined by this relation.
    *
    * @return  Left-hand table AST.
    */
   Aast getLeft()
   {
      return left;
   }
   
   /**
    * Get the right-hand table of the join defined by this relation.
    *
    * @return  Right-hand table AST.
    */
   Aast getRight()
   {
      return right;
   }
   
   /**
    * Get the table which defines the primary key end of this relation, if
    * any. If this method returns <code>null</code>, it indicates a
    * one-to-one relationship between the tables
    *
    * @return  Primary key table AST, or <code>null</code> if a two-way,
    *          one-to-one relation was detected.
    */
   Aast getPrimary()
   {
      return primary;
   }
   
   /**
    * Get the table which defines the foreign key end of this relation, if
    * any. If this method returns <code>null</code>, it indicates a
    * one-to-one relationship between the tables
    *
    * @return  Foreign key table AST, or <code>null</code> if a two-way,
    *          one-to-one relation was detected.
    */
   Aast getForeign()
   {
      if (primary == left)
      {
         return right;
      }
      else if (primary == right)
      {
         return left;
      }
      
      return null;
   }
   
   /**
    * Get the set of field names which define the relation described by this
    * object.
    *
    * @return  Set of common field names.
    */
   Set<String> getKey()
   {
      return key;
   }
   
   /**
    * Get the type of relation; will always return a valid type.
    *
    * @return  Relation type; will always be <code>MANY_TO_ONE</code> or
    *          <code>ONE_TO_ONE</code>.
    *
    * @see     #getInverseType
    */
   int getType()
   {
      return type;
   }
   
   /**
    * Get the inverse type of this relation. Whereas {@link #getType}
    * describes the relation from the point of view of the foreign table
    * (or right table, if no foreign end is defined), this method describes
    * the relation from the point of view of the primary table (or left
    * table, if no primary end is defined).
    *
    * @return  Inverse relation type; will always be <code>ONE_TO_MANY</code>
    *          or <code>ONE_TO_ONE</code>.
    */
   int getInverseType()
   {
      int inverse = -1;
      
      switch (type)
      {
         case MANY_TO_ONE:
            inverse = ONE_TO_MANY;
            break;
         case ONE_TO_ONE:
            inverse = ONE_TO_ONE;
            break;
      }
      
      return inverse;
   }
   
   /**
    * Reports which database(s) is/are affected by this relation as a set
    * of (lowercase) database names. Generally, the returned set will
    * include only a single database name. In the (rare) case of a join
    * across databases, the set will contain two database names.
    *
    * @return  An unmodifiable set of the name(s) of database(s) affected by
    *          this join.
    */
   Set<String> getDatabases()
   {
      return Collections.unmodifiableSet(databases);
   }
}
